// ---------------------------------------------------------------------------
//
//  This file is part of PermLib.
//
// Copyright (c) 2009-2010 Thomas Rehn <thomas@carmen76.de>
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// ---------------------------------------------------------------------------


#include "permutation.h"
#include "permutationword.h"
#include "bsgs.h"

#include "transversal/shallow_schreier_tree_transversal.h"
#include "transversal/schreier_tree_transversal.h"
#include "transversal/explicit_transversal.h"
#include "construct/schreier_sims_construction.h"
#include "change/conjugating_base_change.h"
#include "change/simple_base_change.h"
#include "change/new_base_change.h"
#include "change/random_base_transpose.h"
#include "change/deterministic_base_transpose.h"

#include "construct/random_schreier_sims_construction.h"
#include "generator/bsgs_random_generator.h"

#include <iostream>
#include <fstream>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include "timer.h"

using namespace boost;
using namespace permlib;

/*
template<class RandomAccessIterator, class OutputIterator>
void random_sample(RandomAccessIterator begin, RandomAccessIterator end, uint k, OutputIterator outBegin) {
	std::random_shuffle(begin, end);
	for (uint i = 0; i < k; ++i) {
		*(outBegin++) = begin++;
	}
}
*/

int numberOfRuns = 0;

template<class PERM, class TRANSVERSAL, class BASETRANSPOSE, class BASECHANGE>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    std::cout << caseName << std::endl;

	std::ifstream file;
	file.open(filenameGroup);
	if (!file.is_open()) {
		std::cerr << "opening " << filenameGroup << " failed" << std::endl;
		return;
	}

	PERMlist groupGenerators;
	boost::uint64_t order = 0;
	file >> order;
	
	unsigned int my_n;
	file >> my_n;

	std::string line;
	while (!file.eof()) {
		std::getline(file, line);
		if (line.length() < 1)
			continue;
		boost::shared_ptr<PERM> gen(new PERM(my_n, line));
		groupGenerators.push_back(gen);
	}
	file.close();

	
	SchreierSimsConstruction<PERM, TRANSVERSAL> ssc(my_n);
	BSGS<PERM, TRANSVERSAL> bsgs = ssc.construct(groupGenerators.begin(), groupGenerators.end());

	const double t1 = processUserTime();
	//print_info(bsgs);
	
	file.open(filenameSets);
	if (!file.is_open()) {
		std::cerr << "opening " << filenameSets << " failed" << std::endl;
		return;
	}
	
	uint setSize;
	file >> setSize;

	BASECHANGE bc(bsgs);
	int count = 0;
	
	while (!file.eof() && ++count <= numberOfRuns) {
		std::getline(file, line);
		if (line.length() < 1)
			continue;
		std::vector<ulong> newBase(setSize);
		std::vector<std::string> SplitVec;
		split( SplitVec, line, is_any_of(",") );
		
		BOOST_ASSERT(SplitVec.size() == newBase.size());
		
		uint i=0;
		BOOST_FOREACH(std::string &t, SplitVec) {
			newBase[i] = boost::lexical_cast<ulong>(t)-1;
			++i;
		}
		
		BSGS<PERM, TRANSVERSAL> bsgs_copy(bsgs);
		bc.change(bsgs_copy, newBase.begin(), newBase.end(), false);
	}
	file.close();
	PermutationWord::clearStorage();

	const double t2 = processUserTime();
	std::cout << (t2 - t1) << " " << count-1 << " " << bc.m_statTranspositions << " " << bc.m_statScheierGeneratorsConsidered << std::endl;
	//std::cout << "Total: " << (t3-t1) << "; construction: " << (t2-t1) << std::endl;
	//std::cout << "Memory1: " << mem1 << "; Memory2: " << mem2 << std::endl;
}

template<class PERM, class TRANS, class BASETRANSPOSE>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,TRANS,BASETRANSPOSE, ConjugatingBaseChange<PERM,TRANS,BASETRANSPOSE> >(filenameGroup, filenameSets, caseName + " -- ConjugatingBaseChange");
    runCase<PERM,TRANS,BASETRANSPOSE, SimpleBaseChange<PERM,TRANS,BASETRANSPOSE> >(filenameGroup, filenameSets, caseName + " -- SimpleBaseChange");
    runCase<PERM,TRANS,BASETRANSPOSE, NewBaseChange<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- NewBaseChange");
}

template<class PERM, class TRANS>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,TRANS,RandomBaseTranspose<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- RandomBaseTranspose");
    runCase<PERM,TRANS,DeterministicBaseTranspose<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- DeterministicBaseTranspose");
}

int main(int argc, char *argv[]) {
	if (argc < 4) {
        std::cerr << "usage: " << argv[0] << " group setsFile numberOfRuns" << std::endl;
		return -1;
    }
	numberOfRuns = boost::lexical_cast<int>(argv[3]);
	
    header();

    runCase<Permutation, ExplicitTransversal<Permutation> >(argv[1], argv[2], " -- Permutation -- ExplicitTransversal");
    runCase<PermutationWord, SchreierTreeTransversal<PermutationWord> >(argv[1], argv[2], " -- PermutationWord -- SchreierTreeTransversal");
	//runCase<PermutationWord, ShallowSchreierTreeTransversal<PermutationWord> >(argv[1], argv[2], " -- PermutationWord -- ShallowSchreierTreeTransversal");
    
    return 0;
}
