// ---------------------------------------------------------------------------
//
//  This file is part of PermLib.
//
// Copyright (c) 2009-2010 Thomas Rehn <thomas@carmen76.de>
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// ---------------------------------------------------------------------------


#include "permutation.h"
#include "permutationword.h"
#include "bsgs.h"

#include "transversal/shallow_schreier_tree_transversal.h"
#include "transversal/schreier_tree_transversal.h"
#include "transversal/explicit_transversal.h"
#include "construct/schreier_sims_construction.h"
#include "change/conjugating_base_change.h"
#include "change/simple_base_change.h"
#include "change/random_base_transpose.h"
#include "change/deterministic_base_transpose.h"
#include "generator/product_replacement_generator.h"
#include "construct/random_schreier_sims_construction.h"

#include <iostream>
#include <fstream>
#include <climits>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include "timer.h"

using namespace boost;
using namespace permlib;

template<class PERM, class TRANSVERSAL, class BASECONSTRUCTION>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    const uint runsPerGroup = 10;
	std::cout << caseName << " (" << runsPerGroup << ")" << std::endl;

	std::ifstream file;
	file.open(filenameGroup);
	if (!file.is_open()) {
		std::cerr << "opening " << filenameGroup << " failed" << std::endl;
		return;
	}

	uint groupNumber = 1;
	std::string line;
	while (!file.eof()) {
		boost::uint64_t order = 0;
		file >> order;
		unsigned int my_n;
		file >> my_n;
		PERMlist groupGenerators;
		std::getline(file, line);
		
		while (true) {
			std::getline(file, line);
			if (!line.compare("----") || !line.compare("."))
				break;
			boost::shared_ptr<PERM> gen(new PERM(my_n, line));
			groupGenerators.push_back(gen);
		}
		
		//boost::scoped_ptr<ProductReplacementGenerator<PERM> > rng2(new ProductReplacementGenerator<PERM>(groupGenerators.begin(), groupGenerators.end()));
		//BASECONSTRUCTION construction(my_n, rng2.get());
		BASECONSTRUCTION construction(my_n);
		
		uint wrongOrderCount = 0;
		uint maxTransversalDepth = 0;
		const uint mem1 = processSize();
		const double t1 = processUserTime();
		for (uint i = 0; i < runsPerGroup; ++i) {
			//bool guaranteedBSGS;
			//BSGS<PERM, TRANSVERSAL> bsgs = construction.construct(groupGenerators.begin(), groupGenerators.end(), guaranteedBSGS);
			BSGS<PERM, TRANSVERSAL> bsgs(construction.construct(groupGenerators.begin(), groupGenerators.end()));
			if (bsgs.order() != order)
				++wrongOrderCount;
			BOOST_FOREACH(const TRANSVERSAL& U, bsgs.U) {
				if (U.m_statMaxDepth > maxTransversalDepth) {
					maxTransversalDepth = U.m_statMaxDepth;
				}
			}
		}
		const double t2 = processUserTime();
		const uint mem2 = processSize();
		PermutationWord::clearStorage();
		
		std::cout << groupNumber << " " << my_n << " " << order << " " << wrongOrderCount << " " << maxTransversalDepth << " " << (t2-t1) << " " << (mem2 - mem1) << std::endl;
		
		if (!line.compare("."))
			break;
		++groupNumber;
	}
	file.close();
}

template<class PERM, class TRANS>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,TRANS,SchreierSimsConstruction<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- SchreierSims");
	//runCase<PERM,TRANS,RandomSchreierSimsConstruction<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- RandomSchreierSims");
}

template<class PERM>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,ExplicitTransversal<PERM> >(filenameGroup, filenameSets, caseName + " -- ExplicitTransversal");
    runCase<PERM,SchreierTreeTransversal<PERM> >(filenameGroup, filenameSets, caseName + " -- SchreierTreeTransversal");
    runCase<PERM,ShallowSchreierTreeTransversal<PERM> >(filenameGroup, filenameSets, caseName + " -- ShallowSchreierTreeTransversal");
}

int main(int argc, char *argv[]) {
	if (argc < 2) {
        std::cerr << "usage: " << argv[0] << " groupsFile" << std::endl;
		return -1;
    }
    
    header();

    runCase<Permutation>(argv[1], 0, " -- Permutation");
    runCase<PermutationWord>(argv[1], argv[2], " -- PermutationWord");
    
    return 0;
}
