// ---------------------------------------------------------------------------
//
//  This file is part of PermLib.
//
// Copyright (c) 2009-2010 Thomas Rehn <thomas@carmen76.de>
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote products
//    derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
// OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// ---------------------------------------------------------------------------


#include "permutation.h"
#include "permutationword.h"
#include "bsgs.h"

#include "transversal/shallow_schreier_tree_transversal.h"
#include "transversal/schreier_tree_transversal.h"
#include "transversal/explicit_transversal.h"
#include "construct/schreier_sims_construction.h"
#include "change/conjugating_base_change.h"
#include "change/simple_base_change.h"
#include "change/new_base_change.h"
#include "change/random_base_transpose.h"
#include "change/deterministic_base_transpose.h"

#include "search/classic/set_stabilizer_search.h"
#include "search/partition/set_stabilizer_search.h"

#include <iostream>
#include <fstream>
#include <climits>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include "timer.h"

using namespace boost;
using namespace permlib;

int numberOfRuns = 0;

template<class PERM, class TRANSVERSAL, class BASETRANSPOSE, class BASECHANGE, class SEARCH>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName, bool baseChangeRequired, std::list<boost::uint64_t>& groupOrders, uint dcm2Pruning) {
    std::cout << caseName << std::endl;

	std::ifstream file;
	file.open(filenameGroup);
	if (!file.is_open()) {
		std::cerr << "opening " << filenameGroup << " failed" << std::endl;
		return;
	}

	PERMlist groupGenerators;
	boost::uint64_t order = 0;
	file >> order;
	
	unsigned int my_n;
	file >> my_n;

	std::string line;
	while (!file.eof()) {
		std::getline(file, line);
		if (line.length() < 1)
			continue;
		boost::shared_ptr<PERM> gen(new PERM(my_n, line));
		groupGenerators.push_back(gen);
	}
	file.close();

	const uint mem1 = processSize();
	double tSearch = 0.0;
	double tConstruct = 0.0;
	
	SchreierSimsConstruction<PERM, TRANSVERSAL> ssc(my_n);
	BSGS<PERM, TRANSVERSAL> bsgs = ssc.construct(groupGenerators.begin(), groupGenerators.end());
	//print_info(bsgs);
	
	file.open(filenameSets);
	if (!file.is_open()) {
		std::cerr << "opening " << filenameSets << " failed" << std::endl;
		return;
	}
	
	uint setSize;
	file >> setSize;

	BASECHANGE bc(bsgs);
	int count = 0;
	boost::uint64_t visited = 0;
	boost::uint64_t prunedDCM = 0;
	boost::uint64_t prunedDCM2 = 0;
	boost::uint64_t prunedConstraint = 0;
	
	while (!file.eof() && count < numberOfRuns) {
		std::getline(file, line);
		if (line.length() < 1)
			continue;
		std::vector<ulong> newBase(setSize);
		std::vector<std::string> SplitVec;
		split( SplitVec, line, is_any_of(",") );
		
		BOOST_ASSERT(SplitVec.size() == newBase.size());
		
		uint i=0;
		BOOST_FOREACH(std::string &t, SplitVec) {
			newBase[i] = boost::lexical_cast<ulong>(t)-1;
			++i;
		}
		
		BSGS<PERM, TRANSVERSAL> bsgs_copy(bsgs);
		const double t1 = processUserTime();
		
		if (baseChangeRequired) {
			bc.change(bsgs_copy, newBase.begin(), newBase.end());
		}
		
		//SetwiseStabilizerPredicate<PERM> stabPred(newBase.begin(), newBase.end());
		
		SEARCH rbase(bsgs_copy, dcm2Pruning);
		//rbase.construct(stabPred);
		rbase.construct(newBase.begin(), newBase.end());
		
		const double t1b = processUserTime();
		
		BSGS<PERM,TRANSVERSAL> K(bsgs_copy.n);
		rbase.search(K);
		const double t2 = processUserTime();
		tSearch += t2 - t1;
		tConstruct += t1b - t1;
		
		visited += rbase.m_statNodesVisited;
		prunedDCM += rbase.m_statNodesPrunedCosetMinimality;
		prunedDCM2 += rbase.m_statNodesPrunedCosetMinimality2;
		prunedConstraint += rbase.m_statNodesPrunedChildRestriction;
		
		groupOrders.push_back(K.order());
		++count;
		DEBUG(std::cout << "K = " << K << std::endl;)
		//std::cout << "stats: " << rbase.m_statNodesVisited << " / " << rbase.m_statNodesPrunedCosetMinimality << std::endl;
	}
	file.close();

	const uint mem2 = processSize();
	std::cout << (tSearch) << " " << tConstruct << " " << count << " " << (mem2-mem1) << " " << visited << " " <<
		prunedDCM << " " << prunedDCM2 << " " << prunedConstraint << std::endl;
	//std::cout << "Total: " << (tSearch) << std::endl;
	//std::cout << "Memory1: " << mem1 << "; Memory2: " << mem2 << std::endl;
}

template<class PERM, class TRANS, class BASETRANSPOSE, class BASECHANGE>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
	std::list<boost::uint64_t> ordersRbase;
	std::list<boost::uint64_t> ordersRbase2;
	std::list<boost::uint64_t> ordersBacktrack;
	std::list<boost::uint64_t> ordersBacktrack2;
	runCase<PERM,TRANS,BASETRANSPOSE,BASECHANGE, partition::SetStabilizerSearch<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- RBase", true, ordersRbase, 0);
	runCase<PERM,TRANS,BASETRANSPOSE, ConjugatingBaseChange<PERM,TRANS,BASETRANSPOSE>, classic::SetStabilizerSearch<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- Backtrack", true, ordersBacktrack, 0);
	runCase<PERM,TRANS,BASETRANSPOSE,BASECHANGE, partition::SetStabilizerSearch<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- RBase(DCM2)", true, ordersRbase2, 1);
	runCase<PERM,TRANS,BASETRANSPOSE, ConjugatingBaseChange<PERM,TRANS,BASETRANSPOSE>, classic::SetStabilizerSearch<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- Backtrack(DCM2)", true, ordersBacktrack2, 1);
	
	std::list<boost::uint64_t>::const_iterator it1, it2, it3, it4;
	int i = 1;
	boost::uint64_t minSize = static_cast<boost::uint64_t>(-1);
	double avgSize = 0.0;
	for (it1 = ordersRbase.begin(), it2 = ordersBacktrack.begin(),
		 it3 = ordersRbase2.begin(), it4 = ordersBacktrack2.begin();
		 it1 != ordersRbase.end();
		 ++it1, ++it2, ++it3, ++it4, ++i) {
		if (*it1 != *it2 || *it2 != *it3 || *it3 != *it4) {
			std::cout << "group sizes mismatch at " << i << " : " << *it1 << " <> " << *it2 << " <> " << *it3 << " <> " << *it4 << std::endl;
		}
		if (minSize > *it1)
			minSize = *it1;
		avgSize += *it1;
	}

	//avgSize /= ordersRbase.size();
	//std::cout << "min |K| = " << minSize << "; avg |K| = " << avgSize << std::endl;
}

template<class PERM, class TRANS, class BASETRANSPOSE>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,TRANS,BASETRANSPOSE, ConjugatingBaseChange<PERM,TRANS,BASETRANSPOSE> >(filenameGroup, filenameSets, caseName + " -- ConjugatingBaseChange");
	//runCase<PERM,TRANS,BASETRANSPOSE, NewBaseChange<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- NewBaseChange");
}

template<class PERM, class TRANS>
void runCase(const char* filenameGroup, const char* filenameSets, std::string caseName) {
    runCase<PERM,TRANS,RandomBaseTranspose<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- RandomBaseTranspose");
	//runCase<PERM,TRANS,DeterministicBaseTranspose<PERM,TRANS> >(filenameGroup, filenameSets, caseName + " -- DeterministicBaseTranspose");
}

int main(int argc, char *argv[]) {
	if (argc < 4) {
        std::cerr << "usage: " << argv[0] << " group setsFile numberOfRuns" << std::endl;
		return -1;
    }
    
	numberOfRuns = boost::lexical_cast<int>(argv[3]);
	
    header();

	typedef ExplicitTransversal<Permutation> TRANS1;
    runCase<Permutation,TRANS1,RandomBaseTranspose<Permutation,TRANS1> >(argv[1], argv[2], " -- Permutation -- ExplicitTransversal -- RandomBaseTranspose");
	typedef SchreierTreeTransversal<PermutationWord> TRANS2;
	runCase<PermutationWord, TRANS2,DeterministicBaseTranspose<PermutationWord,TRANS2> >(argv[1], argv[2], " -- PermutationWord -- SchreierTreeTransversal -- DeterministicBaseTranspose");
	
	//runCase<PermutationWord, ShallowSchreierTreeTransversal<PermutationWord> >(argv[1], argv[2], " -- PermutationWord -- ShallowSchreierTreeTransversal");
    
    return 0;
}
